/* $Id$ */
/**************************************************************************
 *
 *  FILE            network.c
 *  MODULE OF       vidgrabber
 *
 *  DESCRIPTION     
 *
 *  WRITTEN BY      Sverre H. Huseby <sverrehu@ifi.uio.no>
 *
 **************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <shhmsg.h>
#include <shhnet.h>

#include "blockimg.h"
#include "camera.h"
#include "shhsched.h"
#include "vidgrabber.h"
#include "network.h"

/**************************************************************************
 *                                                                        *
 *                       P R I V A T E    D A T A                         *
 *                                                                        *
 **************************************************************************/

/* need some extra space */
static unsigned char netOutPacket[MAX_PACKET + BLOCK_WIDTH * BLOCK_HEIGHT];
static int    netOutIndex;
static int    netOutImageNumber;

static char   netRemoteSystem[81] = "";
static char   netRemoteHost[81];
static char   netRemoteTcpPort[81];
static int    netRemoteTcp = -1;

static SNAddr netRemoteUdpAddr;
static SNPort netRemoteUdpPort;
static int    netRemoteUdp = -1;

SchedId netTcpReadyId;



/**************************************************************************
 *                                                                        *
 *                   P R I V A T E    F U N C T I O N S                   *
 *                                                                        *
 **************************************************************************/

static void
netTcpInput(void *clientData, int fd)
{
    unsigned char buff[128];
    char          name[100];
    int           n, type;
    SNPort        dummy;

    if (tcpRead(fd, buff, sizeof(buff), &n) != SN_OK || n == 0)
	msgFatal("got error or empty packet, assuming broken connection\n");

    type = buff[0];
    switch (type) {
      case PACK_HAVE_FEED:
	msgFatal("another camera provider running\n");
	break;
      case PACK_WELCOME_PORT:
	inGetPeerAddrPort(fd, &netRemoteUdpAddr, &dummy);
	inHostAddrToName(netRemoteUdpAddr, name, sizeof(name));
	netRemoteUdpPort = ((int) buff[1] << 8) | buff[2];
	udpClientSock(&netRemoteUdp);
	msgMessage("connected to %s, sending UDP data to port %u\n",
		   name, (unsigned) netRemoteUdpPort);
	break;
      case PACK_START_GRAB:
	msgMessage("starting grab on request\n");
	vidStartGrab();
	break;
      case PACK_STOP_GRAB:
	msgMessage("stopping grab on request\n");
	vidStopGrab();
	break;
      default:
	msgError("unknown packet received (%u)\n", (unsigned) type);
    }
}

static void
netFlush(void)
{
    int n;

    if (!netOutIndex)
	return;
    if (netRemoteUdp >= 0) {
	udpWriteTo(netRemoteUdp, netOutPacket, netOutIndex,
		   netRemoteUdpAddr, netRemoteUdpPort, &n);
	if (n < netOutIndex)
	    msgError("partial packet sent\n");
    }
    netOutIndex = 0;
}

static void
netSendBlock(Block *block, int x, int y)
{
#if GRAY_BITS == 4
    unsigned char          pix;
#endif
    int                    n, origOutIndex, nBytes;
    unsigned char          buff[BLOCK_WIDTH * BLOCK_HEIGHT];
    register unsigned char *s, *d;

    s = block->data;
    d = buff;
#if GRAY_BITS == 4
    nBytes = (BLOCK_WIDTH * BLOCK_HEIGHT) / 2;
    for (n = nBytes; n; n--) {
	pix = *s++ & 0xF0;
	*d++ = pix | (*s++ >> 4);
    }
#else
    nBytes = BLOCK_WIDTH * BLOCK_HEIGHT;
    for (n = nBytes; n; n--)
	*d++ = *s++ >> (8 - GRAY_BITS);
#endif

    if (netOutIndex + 3 + BLOCK_WIDTH * BLOCK_HEIGHT > MAX_PACKET)
	netFlush();
    origOutIndex = netOutIndex;
    for (;;) {
	netOutPacket[netOutIndex++] = PACK_BLOCK;
	netOutPacket[netOutIndex++] = (unsigned char) x;
	netOutPacket[netOutIndex++] = (unsigned char) y;
	memcpy(netOutPacket + netOutIndex, buff, nBytes);
	netOutIndex += nBytes;
	if (netOutIndex > MAX_PACKET) {
	    netOutIndex = origOutIndex;
	    netFlush();
	    origOutIndex = netOutIndex;
	} else
	    break;
    }
    block->measureLastSend = block->measure;
    block->lastImageNumber = netOutImageNumber;
}

static void
netCallRemote(void)
{
    msgMessage("calling proxy at %s:%s\n", netRemoteHost, netRemoteTcpPort);
    tcpClientOpen(netRemoteHost, netRemoteTcpPort, &netRemoteTcp);
    netTcpReadyId
	= schedAddDesc(SCHED_READ_READY, netRemoteTcp, netTcpInput, NULL);
}

static void
netGetHostAndPort(const char *system, char *host, char *port)
{
    const char *p;

    if ((p = strchr(system, ':')) == NULL) {
	strcpy(host, system);
	sprintf(port, "%d", DEFAULT_CAMERA_PORT);
    } else {
	if (p == system)
	    strcpy(host, "localhost");
	else
	    strncpy(host, system, p - system);
	strcpy(port, p + 1);
	if (!strlen(port))
	    sprintf(port, "%d", DEFAULT_CAMERA_PORT);
    }
}



/**************************************************************************
 *                                                                        *
 *                    P U B L I C    F U N C T I O N S                    *
 *                                                                        *
 **************************************************************************/

void
netSetRemoteSystem(const char *system)
{
    strcpy(netRemoteSystem, system);
}

void
netInit(void)
{
    netOutIndex = 0;
    netOutImageNumber = RESEND_FREQUENCY;

    netGetHostAndPort(netRemoteSystem, netRemoteHost, netRemoteTcpPort);
    netCallRemote();
}

void
netFinish(void)
{
    if (netRemoteTcp >= 0) {
	tcpClose(netRemoteTcp);
	netRemoteTcp = -1;
    }
    if (netRemoteUdp >= 0) {
	tcpClose(netRemoteUdp);
	netRemoteUdp = -1;
    }
    schedRemoveDesc(netTcpReadyId);
}

void
netSendImage(BlockImage *blockImage)
{
    int x, y, diff;
    Block *block;

    for (y = 0; y < VER_BLOCKS; y++)
	for (x = 0; x < HOR_BLOCKS; x++) {
	    block = &(blockImage->block[x][y]);
	    if (netOutImageNumber - block->lastImageNumber
		>= RESEND_FREQUENCY) {
		netSendBlock(block, x, y);
	    } else {
		diff = block->measure - block->measureLastSend;
		if (diff < 0)
		    diff = -diff;
		if (diff > MEASURE_MAX_DELTA)
		    netSendBlock(block, x, y);
	    }
	}
    netFlush();
    ++netOutImageNumber;
}
